/*
 * File registry and opening functionality.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "directory_resource.h"
#include "file_pager.h"
#include "file_paging.h"

#include <systypes/fcntl.h>



/* The file paging coordinator provides a memory page collection to the page
   mappers it creates. */

FilePaging::FilePaging(Pages *pages)
: _pages(pages)
{
}



/* Convert opening flags to map-compatible paging flags. */

map_flags_t FilePaging::get_flags(flags_t flags)
{
    return flags & (O_WRONLY | O_RDWR) ? L4RE_DS_MAP_FLAG_RW : L4RE_DS_MAP_FLAG_RO;
}



/* Obtain a provider for the 'fileid' or register a new one. */

long FilePaging::get_directory_provider(FileOpening *opening, const char *path,
                                        flags_t flags, fileid_t fileid,
                                        DirectoryProvider **directory_provider)
{
    /* Obtain any registered provider. */

    Accountable *provider = get(fileid);

    if (provider != NULL)
    {
        *directory_provider = dynamic_cast<DirectoryProvider *>(provider);

        if ((*directory_provider) != NULL)
            return L4_EOK;
        else
            return -L4_EIO;
    }

    /* Make an accessor and a provider to encapsulate it. */

    DirectoryAccessor *accessor;
    long err = opening->make_directory_accessor(path, flags, fileid, &accessor);

    if (err)
        return err;

    *directory_provider = new DirectoryProvider(accessor);

    /* Register the provider. */

    set(fileid, *directory_provider);
    return L4_EOK;
}

/* Obtain a provider for the 'fileid' or register a new one. */

long FilePaging::get_file_provider(FileOpening *opening, const char *path,
                                   flags_t flags, fileid_t fileid,
                                   FileProvider **file_provider)
{
    /* Obtain any registered provider. */

    Accountable *provider = get(fileid);

    if (provider != NULL)
    {
        *file_provider = dynamic_cast<FileProvider *>(provider);

        if ((*file_provider) != NULL)
            return L4_EOK;
        else
            return -L4_EIO;
    }

    /* Make an accessor, page mapper, and a provider to encapsulate them. */

    Accessor *accessor;
    long err = opening->make_accessor(path, flags, fileid, &accessor);

    if (err)
        return err;

    PageMapper *mapper = new PageMapper(accessor, _pages);
    *file_provider = new FileProvider(mapper);

    /* Register the provider. */

    set(fileid, *file_provider);
    return L4_EOK;
}



/* Return a directory resource initialised with a provider. */

long FilePaging::get_directory(FileOpening *opening, const char *path,
                               flags_t flags, fileid_t fileid,
                               Resource **resource)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain any existing provider registered for the object, or make a new
       provider. */

    DirectoryProvider *provider;
    long err = get_directory_provider(opening, path, flags, fileid, &provider);

    if (err)
        return err;

    /* Initialise the resource with the provider and a reference to this object
       for detaching from the provider. */

    provider->attach();
    *resource = new DirectoryResource(fileid, provider, this);
    return L4_EOK;
}

/* Return a pager initialised with a provider, page mapper and accessor. */

long FilePaging::get_pager(FileOpening *opening, const char *path,
                           flags_t flags, fileid_t fileid, Pager **pager)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Obtain any existing provider registered for the file, or make a new
       provider. */

    FileProvider *provider;
    long err = get_file_provider(opening, path, flags, fileid, &provider);

    if (err)
        return err;

    /* Initialise the pager with the provider and a reference to this object for
       detaching from the provider. */

    provider->attach();
    *pager = new FilePager(fileid, provider, get_flags(flags), this);
    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
